﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using HeuristicLab.Algorithms.Bandits;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization {
  public class ContextualMctsSampler {
    private class TreeNode {
      public string ident;
      public int randomTries;
      public int policyTries;
      public TreeNode[] children;
      public bool done = false;

      public TreeNode(string id) {
        this.ident = id;
      }

      public override string ToString() {
        return string.Format("Node({0} tries: {1}, done: {2})", ident, randomTries + policyTries, done);
      }
    }


    public event Action<string, double> FoundNewBestSolution;
    public event Action<string, double> SolutionEvaluated;

    private readonly int maxLen;
    private readonly IProblem problem;
    private readonly Random random;
    private readonly int randomTries;
    private readonly Func<Random, int, IPolicy> policyFactory;
    private readonly Dictionary<string, IPolicy> policyForState;

    private List<Tuple<TreeNode, int, string>> updateChain;
    private TreeNode rootNode;

    public int treeDepth;
    public int treeSize;

    public ContextualMctsSampler(IProblem problem, int maxLen, Random random) :
      this(problem, maxLen, random, 10, (rand, numActions) => new EpsGreedyPolicy(rand, numActions, 0.1)) {

    }

    public ContextualMctsSampler(IProblem problem, int maxLen, Random random, int randomTries, Func<Random, int, IPolicy> policyFactory) {
      this.maxLen = maxLen;
      this.problem = problem;
      this.random = random;
      this.randomTries = randomTries;
      this.policyFactory = policyFactory;
      this.policyForState = new Dictionary<string, IPolicy>();
    }

    public void Run(int maxIterations) {
      double bestQuality = double.MinValue;
      InitPolicies(problem.Grammar);
      for (int i = 0; !rootNode.done && i < maxIterations; i++) {
        var sentence = SampleSentence(problem.Grammar).ToString();
        var quality = problem.Evaluate(sentence) / problem.GetBestKnownQuality(maxLen);
        Debug.Assert(quality >= 0 && quality <= 1.0);
        DistributeReward(quality);

        RaiseSolutionEvaluated(sentence, quality);

        if (quality > bestQuality) {
          bestQuality = quality;
          RaiseFoundNewBestSolution(sentence, quality);
        }
      }

      // clean up
      InitPolicies(problem.Grammar); GC.Collect();
    }

    public void PrintStats() {
      var n = rootNode;
      Console.WriteLine("depth: {0,5} size: {1,10} root tries {2,10}", treeDepth, treeSize, rootNode.policyTries + rootNode.randomTries);
      foreach (var p in policyForState.OrderBy(p => p.Key.Length).ThenBy(p => p.Key)) {
        Console.Write("{0,10} {1,20}", p.Key, p.Value);
        p.Value.PrintStats();
      }
      //while (n.policy != null) {
      //  Console.WriteLine();
      //  Console.WriteLine("{0,5}->{1,-50}", n.ident, string.Join(" ", n.children.Select(ch => string.Format("{0,4}", ch.ident))));
      //  Console.WriteLine("{0,5}  {1,-50}", string.Empty, string.Join(" ", n.children.Select(ch => string.Format("{0,4}", ch.randomTries + ch.policyTries))));
      //  //n.policy.PrintStats();
      //  n = n.children.OrderByDescending(c => c.policyTries).First();
      //}
      Console.ReadLine();
    }

    private void InitPolicies(IGrammar grammar) {
      this.updateChain = new List<Tuple<TreeNode, int, string>>();
      policyForState.Clear();
      rootNode = new TreeNode(grammar.SentenceSymbol.ToString());
      treeDepth = 0;
      treeSize = 0;
    }

    private Sequence SampleSentence(IGrammar grammar) {
      updateChain.Clear();
      var startPhrase = new Sequence(grammar.SentenceSymbol);
      return CompleteSentence(grammar, startPhrase);
    }

    private Sequence CompleteSentence(IGrammar g, Sequence phrase) {
      if (phrase.Length > maxLen) throw new ArgumentException();
      if (g.MinPhraseLength(phrase) > maxLen) throw new ArgumentException();
      TreeNode n = rootNode;
      bool done = phrase.IsTerminal;
      int selectedAltIdx = -1;
      var curDepth = 0;
      while (!done) {
        var terminalSequence = phrase.Subsequence(0, phrase.FirstNonTerminalIndex).ToString();

        char nt = phrase.FirstNonTerminal;

        int maxLenOfReplacement = maxLen - (phrase.Length - 1); // replacing aAb with maxLen 4 means we can only use alternatives with a minPhraseLen <= 2
        Debug.Assert(maxLenOfReplacement > 0);

        var alts = g.GetAlternatives(nt).Where(alt => g.MinPhraseLength(alt) <= maxLenOfReplacement);

        if (n.randomTries < randomTries) {
          n.randomTries++;

          treeDepth = Math.Max(treeDepth, curDepth);

          return g.CompleteSentenceRandomly(random, phrase, maxLen);
        } else if (n.randomTries == randomTries && n.children == null) {
          n.children = alts.Select(alt => new TreeNode(alt.ToString())).ToArray(); // create a new node for each alternative

          treeSize += n.children.Length;
        }

        IPolicy policy = GetPolicyForTerminalSequence(terminalSequence, alts.Count());
        n.policyTries++;
        // => select using bandit policy
        selectedAltIdx = policy.SelectAction();
        if (selectedAltIdx == -1) {
          // oops
          n.done = true;
          Console.Write("*");
          return g.CompleteSentenceRandomly(random, phrase, maxLen);
        }
        Sequence selectedAlt = alts.ElementAt(selectedAltIdx);
        updateChain.Add(Tuple.Create(n, selectedAltIdx, terminalSequence)); // context is terminal sequence so far

        // replace nt with alt
        phrase.ReplaceAt(phrase.FirstNonTerminalIndex, 1, selectedAlt);


        curDepth++;

        done = phrase.IsTerminal;
        // prepare for next iteration
        n = n.children[selectedAltIdx];
        Debug.Assert(!n.done);
      } // while

      // the last node is a leaf node (sentence is done), so we never need to visit this node again
      n.done = true;

      treeDepth = Math.Max(treeDepth, curDepth);
      return phrase;
    }

    private void DistributeReward(double reward) {
      // iterate in reverse order (bottom up)
      updateChain.Reverse();

      foreach (var e in updateChain) {
        var node = e.Item1;
        var action = e.Item2;
        var sequence = e.Item3;
        var policy = GetPolicyForTerminalSequence(sequence, -1);
        //policy.UpdateReward(action, reward / updateChain.Count);
        policy.UpdateReward(action, reward);

        if (node.children[action].done) policy.DisableAction(action);
        if (node.children.All(c => c.done)) node.done = true;
      }
    }

    private IPolicy GetPolicyForTerminalSequence(string sequence, int numAlts) {
      IPolicy p;
      sequence = problem.Hash(sequence);
      if (!policyForState.TryGetValue(sequence, out p)) {
        p = policyFactory(random, numAlts);
        policyForState.Add(sequence, p);
      }
      return p;
    }

    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
    private void RaiseFoundNewBestSolution(string sentence, double quality) {
      var handler = FoundNewBestSolution;
      if (handler != null) handler(sentence, quality);
    }
  }
}
